/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  linux.c -- Nasty, non-portable, linux specific stuff which changes
             from kernel release to kernel release. ie the transparent 
             proxy calls.

***************************************/

#include <sys/utsname.h>
#include "common.h"

#if HAVE_LINUX_NETFILTER_IPV4_H
# include <linux/netfilter_ipv4.h>
#endif

void get_kernel(void);

static enum {
	INIT,
	LINUX_2_2,
	LINUX_2_4,
	OTHER
} kernel = INIT;


/* ------------------------------------------------------------- **
**  Get the original destination address of a transparently proxied
**  socket.
**  ------------------------------------------------------------- */
int get_orig_dest(int fd, struct sockaddr *addr)
{
	socklen_t len;

	if(kernel==INIT) get_kernel();

	len=sizeof(*addr);
	switch(kernel){
	case LINUX_2_2:
		return( getsockname(fd, addr, &len) );
	default:
#ifdef SO_ORIGINAL_DST /*Header support for kernel 2.4*/
		return( getsockopt(fd, SOL_IP, SO_ORIGINAL_DST, 
				   addr, &len) );
#endif
		write_log(ERROR, "Running on a kernel we haven't been compiled for. Oooops.");
		return(-1);
	}
}

/* ------------------------------------------------------------- **
**  Get the address of the interface we connect to the client through
**  for putting in our 227 reply. For 2.4 do a getsockname on the
**  control socket. For 2.2 this gives us the orriginal destination of
**  the transparently proxied connection, so we do some nasty hackery
**  instead.
**  ------------------------------------------------------------- */
int get_local_address(struct sockaddr_in *addr)
{
	int sockfd;
        socklen_t len;

	/*If ListenAddress is in the config file then use the address
	 * from that*/
	*addr=config_getaddress();
	if(addr->sin_addr.s_addr!=0) {
		addr->sin_port=0;
		return(0);
	}

        if(kernel==INIT) get_kernel();
 
	switch(kernel){
	case LINUX_2_2:
		/* This piece of code ought to be taken out and shot
		**  (yes - it opens a UDP socket, does a getsockname,
		**  and then closes the socket before anything
		**  happens!) Suggestions for an alternative welcomed */
		
		if( (sockfd = socket(AF_INET, SOCK_DGRAM, 0))==-1) return(-1);

		addr->sin_family      = AF_INET;
		addr->sin_addr        = info->client_control.address.sin_addr;
		addr->sin_port        = htons(12345);
		len = sizeof(*addr);
			
		if(connect(sockfd, addr, len)==-1) {
			close(sockfd);
			return(-1);
		}
		
		if(getsockname(sockfd, addr, &len)==-1) {
			close(sockfd);
			return(-1);
		}

		close(sockfd);

		addr->sin_port=0;
		return(0);
	case LINUX_2_4:
	default:
                len=sizeof(*addr);
                return( getsockname(info->client_control.fd, addr, &len) );
	}
}

#if 0
/* ------------------------------------------------------------- **
**  Make a connection to dest which appears to have come from src. Use
**  netfilter snat (through libiptc) on kernel 2.4, and bind() to
**  foreign address on 2.2. If we can't do it transparently, make the
**  connection anyway.
**  ------------------------------------------------------------- */
int transp_connect(struct sockaddr_in dest, struct sockaddr_in src)
{
	/*TODO*/
	return(-1);
}

int tc_close(int fd)
{
	/*TODO*/
	return(-1);
}

/* ------------------------------------------------------------- **
**  Listen in order to intercept any connections that are headed for
**  realdest, on a port from the range specified by use. Use netfilter
**  dnat (through libiptc) on kernel 2.4, or the ipfw setsockopt stuff
**  on 2.2.
**  ------------------------------------------------------------- */
int intercept_listen(struct sockaddr_in intercept, struct sockaddr_in listen_on, socketuse use)
{
	/*TODO*/
	return(-1);
}

int il_close(int fd)
{
	/*TODO*/
	return(-1);
}
#endif

void get_kernel(void)
{
	struct utsname tmp;
	
	uname(&tmp);
	if(!strncmp(tmp.release,"2.2.",4)) kernel=LINUX_2_2;
	else if(!strncmp(tmp.release,"2.4.",4)) kernel=LINUX_2_4;
	else kernel=OTHER;
}
